import threading

__all__ = ["USBCamera", "CameraServer"]

class USBCamera:
    kDefaultCameraName = b"cam0"

    class WhiteBalance:
        kFixedIndoor = 3000
        kFixedOutdoor1 = 4000
        kFixedOutdoor2 = 5000
        kFixedFluorescent1 = 5100
        kFixedFlourescent2 = 5200

    def __init__(self, name=None):
        if name is None:
            name = USBCamera.kDefaultCameraName
        self.name = name
        self.id = None
        self.active = False
        self.useJpeg = True

        self.mutex = threading.RLock()

        self.width = 320
        self.height = 240
        self.fps = 30
        self.whiteBalance = "auto"
        self.whiteBalanceValue = None
        self.exposure = "auto"
        self.exposureValue = None
        self.brightness = 50
        self.needSettingsUpdate = True

        self.openCamera()

    def openCamera(self):
        pass

    def closeCamera(self):
        pass

    def startCapture(self):
        self.active = True

    def stopCapture(self):
        self.active = False

    def updateSettings(self):
        pass

    def setFPS(self, fps):
        with self.mutex:
            if fps != self.fps:
                self.needSettingsUpdate = True
                self.fps = fps

    def setSize(self, width, height):
        with self.mutex:
            if width != self.width or height != self.height:
                self.needSettingsUpdate = True
                self.width = width
                self.height = height

    def setBrightness(self, brightness):
        """Set the brightness, as a percentage (0-100).
        """
        with self.mutex:
            if brightness > 100:
                self.brightness = 100
            elif brightness < 0:
                self.brightness = 0
            else:
                self.brightness = brightness
            self.needSettingsUpdate = True

    def getBrightness(self):
        """Get the brightness, as a percentage (0-100).
        """
        with self.mutex:
            return self.brightness

    def setWhiteBalanceAuto(self):
        """Set the white balance to auto.
        """
        with self.mutex:
            self.whiteBalance = "auto"
            self.whiteBalanceValue = None
            self.needSettingsUpdate = True

    def setWhiteBalanceHoldCurrent(self):
        """Set the white balance to hold current.
        """
        with self.mutex:
            self.whiteBalance = "manual"
            self.whiteBalanceValue = None
            self.needSettingsUpdate = True

    def setWhiteBalanceManual(self, value):
        """Set the white balance to manual, with specified color temperature.
        """
        with self.mutex:
            self.whiteBalance = "manual"
            self.whiteBalanceValue = value
            self.needSettingsUpdate = True

    def setExposureAuto(self):
        """Set the exposure to auto aperature.
        """
        with self.mutex:
            self.exposure = "auto"
            self.exposureValue = None
            self.needSettingsUpdate = True

    def setExposureHoldCurrent(self):
        """Set the exposure to hold current.
        """
        with self.mutex:
            self.exposure = "manual"
            self.exposureValue = None
            self.needSettingsUpdate = True

    def setExposureManual(self, value):
        """Set the exposure to manual, as a percentage (0-100).
        """
        with self.mutex:
            self.exposure = "manual"
            if value > 100:
                self.exposureValue = 100
            elif value < 0:
                self.exposureValue = 0
            else:
                self.exposureValue = value
            self.needSettingsUpdate = True

    def getImage(self, image):
        with self.mutex:
            if self.needSettingsUpdate or self.useJpeg:
                self.needSettingsUpdate = False
                self.useJpeg = False
                self.updateSettings()
        raise NotImplementedError

    def getImageData(self, data, maxsize):
        with self.mutex:
            if self.needSettingsUpdate or not self.useJpeg:
                self.needSettingsUpdate = False
                self.useJpeg = True
                self.updateSettings()
        return 0

class CameraServer:
    kPort = 1180
    kSize640x480 = 0
    kSize320x240 = 1
    kSize160x120 = 2

    server = None

    @staticmethod
    def getInstance():
        if CameraServer.server is None:
            CameraServer.server = CameraServer()
        return CameraServer.server

    def __init__(self):
        self.mutex = threading.RLock()
        self.ready = threading.Event()
        self.quality = 50
        self.camera = None

    def setImage(self, image):
        pass

    def startAutomaticCapture(self, camera):
        """Start automatically capturing images to send to the dashboard.

        You should call this method to just see a camera feed on the dashboard
        without doing any vision processing on the roboRIO. {@link #setImage}
        shouldn't be called after this is called.

        :param camera: The camera interface (e.g. a USBCamera instance)
        """
        if self.camera is not None:
            return

        self.camera = camera
        self.camera.startCapture()

    def isAutoCaptureStarted(self):
        """check if auto capture is started
        """
        with self.mutex:
            return self.camera is not None

    def setSize(self, size):
        if size == self.kSize160x120:
            self.camera.setSize(160, 120)
        elif size == self.kSize320x240:
            self.camera.setSize(320, 240)
        elif size == self.kSize640x480:
            self.camera.setSize(640, 480)
        else:
            return

    def setQuality(self, quality):
        """Set the quality of the compressed image sent to the dashboard

        :param quality: The quality of the JPEG image, from 0 to 100
        """
        with self.mutex:
            if quality > 100:
                self.quality = 100
            elif quality < 0:
                self.quality = 0
            else:
                self.quality = quality

    def getQuality(self):
        """Get the quality of the compressed image sent to the dashboard

        :returns: The quality, from 0 to 100
        """
        with self.mutex:
            return self.quality




#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .button import Button

__all__ = ["NetworkButton"]

class NetworkButton(Button):
    def __init__(self, table, field):
        from networktables import NetworkTable
        if isinstance(table, NetworkTable):
            self.table = table
        else:
            self.table = NetworkTable.getTable(table)
        self.field = field

    def get(self):
        return self.table.isConnected() and self.table.getBoolean(self.field, False)



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

from .livewindow import LiveWindow
from .safepwm import SafePWM

__all__ = ["Victor"]

class Victor(SafePWM):
    """
        VEX Robotics Victor 888 Speed Controller via PWM
        
        The Vex Robotics Victor 884 Speed Controller can also be used with this
        class but may need to be calibrated per the Victor 884 user manual.
        
        .. note ::

            The Victor uses the following bounds for PWM values.  These
            values were determined empirically and optimized for the Victor
            888. These values should work reasonably well for Victor 884
            controllers also but if users experience issues such as
            asymmetric behaviour around the deadband or inability to saturate
            the controller in either direction, calibration is recommended.
            The calibration procedure can be found in the Victor 884 User
            Manual available from VEX Robotics:
            http://content.vexrobotics.com/docs/ifi-v884-users-manual-9-25-06.pdf
            
            - 2.027ms = full "forward"
            - 1.525ms = the "high end" of the deadband range
            - 1.507ms = center of the deadband range (off)
            - 1.49ms = the "low end" of the deadband range
            - 1.026ms = full "reverse"
        
        .. not_implemented: initVictor
    """
    
    def __init__(self, channel):
        """Constructor.

        :param channel: The PWM channel that the Victor is attached to. 0-9 are on-board, 10-19 are on the MXP port
        :type  channel: int
        """
        super().__init__(channel)
        self.setBounds(2.027, 1.525, 1.507, 1.49, 1.026)
        self.setPeriodMultiplier(self.PeriodMultiplier.k2X)
        self.setRaw(self.centerPwm)
        self.setZeroLatch()

        LiveWindow.addActuatorChannel("Victor", self.getChannel(), self)
        hal.HALReport(hal.HALUsageReporting.kResourceType_Victor,
                      self.getChannel())

    def set(self, speed, syncGroup=0):
        """Set the PWM value.

        The PWM value is set using a range of -1.0 to 1.0, appropriately
        scaling the value for the FPGA.

        :param speed: The speed to set.  Value should be between -1.0 and 1.0.
        :type  speed: float
        :param syncGroup: The update group to add this set to, pending
            updateSyncGroup().  If 0, update immediately.
        """
        self.setSpeed(speed)
        self.feed()

    def get(self):
        """Get the recently set value of the PWM.

        :returns: The most recently set value for the PWM between -1.0 and 1.0.
        :rtype: float
        """
        return self.getSpeed()

    def pidWrite(self, output):
        """Write out the PID value as seen in the PIDOutput base object.

        :param output: Write out the PWM value as was found in the
            :class:`PIDController`.
        :type  output: float
        """
        self.set(output)



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .command import Command

__all__ = ["WaitForChildren"]

class WaitForChildren(Command):
    """This command will only finish if whatever :class:`.CommandGroup` it
    is in has no active children.  If it is not a part of a CommandGroup,
    then it will finish immediately.  If it is itself an active child, then
    the CommandGroup will never end.

    This class is useful for the situation where you want to allow anything
    running in parallel to finish, before continuing in the main CommandGroup
    sequence.
    """

    def isFinished(self):
        return self.getGroup() is None or not self.getGroup().children



import nivision
import logging
import re
import threading
import time

__all__ = ["USBCamera"]

logger = logging.getLogger(__name__)

def getJpegSize(data):
    if data[0] != 0xff or data[1] != 0xd8:
        raise ValueError("invalid image")
    pos = 2
    while True:
        if data[pos] != 0xff:
            raise ValueError("invalid image at pos %d (%x)" % (pos, data[pos]))
        t = data[pos+1]
        if t == 0x01 or t >= 0xd0 and t <= 0xd7: # various
            pos += 2
        elif t == 0xd9: # EOI
            return pos + 2
        elif t == 0xd8: # SOI
            raise ValueError("invalid image")
        elif t == 0xda: # SOS
            len = ((data[pos+2] & 0xff) << 8) | (data[pos+3] & 0xff)
            pos += len + 2
            # Find next marker.  Skip over escaped and RST markers.
            while data[pos] != 0xff or data[pos+1] == 0x00 or (data[pos+1] >= 0xd0 and data[pos+1] <= 0xd7):
                pos += 1
        else: # various
            len = ((data[pos+2] & 0xff) << 8) | (data[pos+3] & 0xff)
            pos += len + 2

class USBCamera:
    kDefaultCameraName = b"cam0"

    ATTR_WB_MODE = b"CameraAttributes::WhiteBalance::Mode"
    ATTR_WB_VALUE = b"CameraAttributes::WhiteBalance::Value"
    ATTR_EX_MODE = b"CameraAttributes::Exposure::Mode"
    ATTR_EX_VALUE = b"CameraAttributes::Exposure::Value"
    ATTR_BR_MODE = b"CameraAttributes::Brightness::Mode"
    ATTR_BR_VALUE = b"CameraAttributes::Brightness::Value"

    class WhiteBalance:
        kFixedIndoor = 3000
        kFixedOutdoor1 = 4000
        kFixedOutdoor2 = 5000
        kFixedFluorescent1 = 5100
        kFixedFlourescent2 = 5200

    reMode = re.compile(b"(?P<width>[0-9]+)\s*x\s*(?P<height>[0-9]+)\s+(?P<format>.*?)\s+(?P<fps>[0-9.]+)\s*fps")

    def __init__(self, name=None):
        if name is None:
            name = USBCamera.kDefaultCameraName
        self.name = name
        self.id = None
        self.active = False
        self.useJpeg = True

        self.mutex = threading.RLock()

        self.width = 320
        self.height = 240
        self.fps = 30
        self.whiteBalance = "auto"
        self.whiteBalanceValue = None
        self.exposure = "auto"
        self.exposureValue = None
        self.brightness = 50
        self.needSettingsUpdate = True

        self.openCamera()

    def openCamera(self):
        with self.mutex:
            for i in range(3):
                try:
                    self.id = nivision.IMAQdxOpenCamera(self.name,
                            nivision.IMAQdxCameraControlModeController)
                except nivision.ImaqDxError as e:
                    if i == 2:
                        raise
                    time.sleep(2.0)
                    continue
                break

    def closeCamera(self):
        with self.mutex:
            if self.id is None:
                return
            nivision.IMAQdxCloseCamera(self.id)
            self.id = None

    def startCapture(self):
        with self.mutex:
            if self.id is None or self.active:
                return
            nivision.IMAQdxConfigureGrab(self.id)
            nivision.IMAQdxStartAcquisition(self.id)
            self.active = True

    def stopCapture(self):
        with self.mutex:
            if self.id is None or not self.active:
                return
            nivision.IMAQdxStopAcquisition(self.id)
            nivision.IMAQdxUnconfigureAcquisition(self.id)
            self.active = False

    def updateSettings(self):
        with self.mutex:
            wasActive = self.active
            # Stop acquistion, close and reopen camera
            if wasActive:
                self.stopCapture()
            if self.id is not None:
                self.closeCamera()
            self.openCamera()

            # Video Mode
            modes, currentMode = nivision.IMAQdxEnumerateVideoModes(self.id)
            foundMode = None
            foundFps = 1000
            for mode in modes:
                m = self.reMode.match(mode.Name)
                if not m:
                    continue
                if int(m.group("width")) != self.width:
                    continue
                if int(m.group("height")) != self.height:
                    continue
                fps = float(m.group("fps"))
                if fps < self.fps:
                    continue
                if fps > foundFps:
                    continue
                isJpeg = m.group("format") in (b"jpeg", b"JPEG")
                if (self.useJpeg and not isJpeg) or (not self.useJpeg and isJpeg):
                    continue
                foundMode = mode
                foundFps = fps
            if foundMode is not None:
                logger.info("found mode %d: %s" % (foundMode.Value, foundMode.Name))
                if foundMode.Value != currentMode.value:
                    nivision.IMAQdxSetAttribute(self.id, nivision.IMAQdxAttributeVideoMode, foundMode)

            # White Balance
            if self.whiteBalance == "auto":
                nivision.IMAQdxSetAttribute(self.id, self.ATTR_WB_MODE, b"Auto")
            else:
                nivision.IMAQdxSetAttribute(self.id, self.ATTR_WB_MODE, b"Manual")
                if self.whiteBalanceValue is not None:
                    nivision.IMAQdxSetAttribute(self.id, self.ATTR_WB_VALUE, self.whiteBalanceValue)

            # Exposure
            if self.exposure == "auto":
                nivision.IMAQdxSetAttribute(self.id, self.ATTR_EX_MODE, b"AutoAperaturePriority")
            else:
                nivision.IMAQdxSetAttribute(self.id, self.ATTR_EX_MODE, b"Manual")
                if self.exposureValue is not None:
                    minv = nivision.IMAQdxGetAttributeMinimum(self.id, self.ATTR_EX_VALUE)
                    maxv = nivision.IMAQdxGetAttributeMaximum(self.id, self.ATTR_EX_VALUE)
                    val = minv + ((maxv - minv) * (self.exposureValue / 100.0))
                    nivision.IMAQdxSetAttribute(self.id, self.ATTR_EX_VALUE, val)

            # Brightness
            nivision.IMAQdxSetAttribute(self.id, self.ATTR_BR_MODE, b"Manual")
            minv = nivision.IMAQdxGetAttributeMinimum(self.id, self.ATTR_BR_VALUE)
            maxv = nivision.IMAQdxGetAttributeMaximum(self.id, self.ATTR_BR_VALUE)
            val = minv + ((maxv - minv) * (self.brightness / 100.0))
            nivision.IMAQdxSetAttribute(self.id, self.ATTR_BR_VALUE, val)

            # Restart acquisition
            if wasActive:
                self.startCapture()

    def setFPS(self, fps):
        with self.mutex:
            if fps != self.fps:
                self.needSettingsUpdate = True
                self.fps = fps

    def setSize(self, width, height):
        with self.mutex:
            if width != self.width or height != self.height:
                self.needSettingsUpdate = True
                self.width = width
                self.height = height

    def setBrightness(self, brightness):
        """Set the brightness, as a percentage (0-100).
        """
        with self.mutex:
            if brightness > 100:
                self.brightness = 100
            elif brightness < 0:
                self.brightness = 0
            else:
                self.brightness = brightness
            self.needSettingsUpdate = True

    def getBrightness(self):
        """Get the brightness, as a percentage (0-100).
        """
        with self.mutex:
            return self.brightness

    def setWhiteBalanceAuto(self):
        """Set the white balance to auto.
        """
        with self.mutex:
            self.whiteBalance = "auto"
            self.whiteBalanceValue = None
            self.needSettingsUpdate = True

    def setWhiteBalanceHoldCurrent(self):
        """Set the white balance to hold current.
        """
        with self.mutex:
            self.whiteBalance = "manual"
            self.whiteBalanceValue = None
            self.needSettingsUpdate = True

    def setWhiteBalanceManual(self, value):
        """Set the white balance to manual, with specified color temperature.
        """
        with self.mutex:
            self.whiteBalance = "manual"
            self.whiteBalanceValue = value
            self.needSettingsUpdate = True

    def setExposureAuto(self):
        """Set the exposure to auto aperature.
        """
        with self.mutex:
            self.exposure = "auto"
            self.exposureValue = None
            self.needSettingsUpdate = True

    def setExposureHoldCurrent(self):
        """Set the exposure to hold current.
        """
        with self.mutex:
            self.exposure = "manual"
            self.exposureValue = None
            self.needSettingsUpdate = True

    def setExposureManual(self, value):
        """Set the exposure to manual, as a percentage (0-100).
        """
        with self.mutex:
            self.exposure = "manual"
            if value > 100:
                self.exposureValue = 100
            elif value < 0:
                self.exposureValue = 0
            else:
                self.exposureValue = value
            self.needSettingsUpdate = True

    def getImage(self, image):
        with self.mutex:
            if self.needSettingsUpdate or self.useJpeg:
                self.needSettingsUpdate = False
                self.useJpeg = False
                self.updateSettings()

            #t0 = time.time()
            nivision.IMAQdxGrab(self.id, image, 1)
            #print("grab: %s" % (time.time() - t0))

    def getImageData(self, data, maxsize):
        with self.mutex:
            if self.needSettingsUpdate or not self.useJpeg:
                self.needSettingsUpdate = False
                self.useJpeg = True
                self.updateSettings()

            #t0 = time.time()
            nivision.IMAQdxGetImageData(self.id, data, maxsize,
                    nivision.IMAQdxBufferNumberModeLast, 0)
            #print("get image data: %s" % (time.time() - t0))
            return getJpegSize(data)




#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
import threading

import logging
logger = logging.getLogger(__name__)

__all__ = ["Preferences"]

class Preferences:
    """Provides a relatively simple way to save important
    values to the RoboRIO to access the next time the RoboRIO is booted.

    This class loads and saves from a file inside the RoboRIO. The user can not
    access the file directly, but may modify values at specific fields which
    will then be saved to the file when :func:`save` is called.

    This class is thread safe.

    This will also interact with :class:`networktables.NetworkTable` by creating a table
    called "Preferences" with all the key-value pairs. To save using
    `NetworkTable`, simply set the boolean at position ~S A V E~ to true.
    Also, if the value of any variable is " in the `NetworkTable`, then
    that represents non-existence in the `Preferences` table.
    """

    # The Preferences table name
    TABLE_NAME = "Preferences"
    # The value of the save field
    SAVE_FIELD = "~S A V E~"
    # The file to save to
    FILE_NAME = "/home/lvuser/wpilib-preferences.ini"
    # The characters to put between a field and value
    VALUE_PREFIX = '="'
    # The characters to put after the value
    VALUE_SUFFIX = '"\n'
    # The newline character
    NEW_LINE = '\n'

    @staticmethod
    def getInstance():
        """Returns the preferences instance.

        :returns: the preferences instance
        """
        if not hasattr(Preferences, "instance"):
            Preferences.instance = Preferences()
        return Preferences.instance

    def __init__(self):
        """Creates a preference class that will automatically read the file in
        a different thread. Any call to its methods will be blocked until the
        thread is finished reading.
        """
        # The actual values (str->str)
        self.values = {}
        # The keys in the order they were read from the file
        self.keylist = []
        # The comments that were in the file sorted by which key they appeared
        # over (str->str)
        self.comments = {}
        # The comment at the end of the file
        self.endComment = ""

        # The semaphore for beginning reads and writes to the file
        self.fileLock = threading.Condition()
        # The semaphore for reading from the table
        self.lock = threading.RLock()

        # We synchronized on fileLock and then wait
        # for it to know that the reading thread has started
        with self.fileLock:
            reader = threading.Thread(target=self._read,
                                      name="Preferences Read")
            reader.start()
            self.fileLock.wait()

        hal.HALReport(hal.HALUsageReporting.kResourceType_Preferences, 0)

    def getKeys(self):
        """:returns: a list of the keys
        """
        with self.lock:
            return [x for x in self.keylist]

    def keys(self):
        """Python style get list of keys.
        """
        with self.lock:
            return [x for x in self.keylist]

    def put(self, key, value):
        """Puts the given value into the given key position

        :param key: the key
        :param value: the value
        """
        if any((c in key) for c in "=\n\r\t[] "):
            raise KeyError("improper preference key '%s'" % key)
        with self.lock:
            if key not in self.values:
                self.keylist.append(key)
            self.values[key] = value
            try:
                from networktables import NetworkTable
                NetworkTable.getTable(self.TABLE_NAME).putString(key, value)
            except ImportError:
                pass

    def putString(self, key, value):
        """Puts the given string into the preferences table.

        The value may not have quotation marks, nor may the key have any
        whitespace nor an equals sign.

        This will NOT save the value to memory between power cycles, to
        do that you must call :func:`save` (which must be used with care)
        at some point after calling this.

        :param key: the key
        :param value: the value
        """
        if '"' in value:
            raise ValueError("Can not put string: '%s' because it contains quotation marks" % value)
        self.put(key, value)

    def putInt(self, key, value):
        """Puts the given int into the preferences table.

        The key may not have any whitespace nor an equals sign.

        This will NOT save the value to memory between power cycles, to
        do that you must call :func:`save` (which must be used with care)
        at some point after calling this.

        :param key: the key
        :param value: the value
        """
        self.put(key, str(value))

    def putFloat(self, key, value):
        """Puts the given float into the preferences table.

        The key may not have any whitespace nor an equals sign.

        This will NOT save the value to memory between power cycles, to
        do that you must call :func:`save` (which must be used with care)
        at some point after calling this.

        :param key: the key
        :param value: the value
        """
        self.put(key, str(value))

    def putBoolean(self, key, value):
        """Puts the given float into the preferences table.

        The key may not have any whitespace nor an equals sign.

        This will NOT save the value to memory between power cycles, to
        do that you must call :func:`save` (which must be used with care)
        at some point after calling this.

        :param key: the key
        :param value: the value
        """
        self.put(key, str(value))

    def __setitem__(self, key, value):
        """Python style setting of key/value."""
        if isinstance(value, str):
            self.putString(key, value)
        else:
            self.put(key, str(value))

    def get(self, key, d=None):
        """Returns the value at the given key.

        :param key: the key
        :param d: the return value if the key doesn't exist (default is None)
        :returns: the value (or d/None if none exists)
        """
        with self.lock:
            return self.values.get(key, d)

    def containsKey(self, key):
        """Returns whether or not there is a key with the given name.

        :param key: the key
        :returns: True if there is a value at the given key
        """
        with self.lock:
            return key in self.values

    def __contains__(self, key):
        """Python style contains key."""
        with self.lock:
            return key in self.values

    def has_key(self, key):
        """Python style contains key."""
        with self.lock:
            return key in self.values

    def remove(self, key):
        """Remove a preference

        :param key: the key
        """
        with self.lock:
            self.values.pop(key, None)
            try:
                self.keylist.remove(key)
            except ValueError:
                pass

    def __delitem__(self, key):
        """Python style preference removal
        """
        with self.lock:
            del self.values[key]
            try:
                self.keylist.remove(key)
            except ValueError:
                raise KeyError(key)

    def getString(self, key, backup):
        """Returns the string at the given key. If this table does not have a
        value for that position, then the given backup value will be returned.

        :param key: the key
        :param backup: the value to return if none exists in the table
        :returns: either the value in the table, or the backup
        """
        return self.get(key, backup)

    def getInt(self, key, backup):
        """Returns the int at the given key. If this table does not have a
        value for that position, then the given backup value will be returned.

        :param key: the key
        :param backup: the value to return if none exists in the table
        :returns: either the value in the table, or the backup
        :raises: ValueError if value cannot be converted to integer
        """
        value = self.get(key)
        if value is None:
            return backup
        return int(value)

    def getFloat(self, key, backup):
        """Returns the float at the given key. If this table does not have a
        value for that position, then the given backup value will be returned.

        :param key: the key
        :param backup: the value to return if none exists in the table
        :returns: either the value in the table, or the backup
        :raises: ValueError if value cannot be converted to integer
        """
        value = self.get(key)
        if value is None:
            return backup
        return float(value)

    def getBoolean(self, key, backup):
        """Returns the boolean at the given key. If this table does not have a
        value for that position, then the given backup value will be returned.

        :param key: the key
        :param backup: the value to return if none exists in the table
        :returns: either the value in the table, or the backup
        :raises: ValueError if value cannot be converted to integer
        """
        value = self.get(key)
        if value is None:
            return backup
        if value.lower() == "true":
            return True
        elif value.lower() == "false":
            return False
        else:
            raise ValueError("invalid literal for boolean: '%s'" % value)

    def save(self):
        """Saves the preferences to a file on the RoboRIO.

        This should NOT be called often. Too many writes can damage the
        RoboRIO's flash memory. While it is ok to save once or twice a match,
        this should never be called every run of
        :func:`IterativeRobot.teleopPeriodic`.

        The actual writing of the file is done in a separate thread. However,
        any call to a get or put method will wait until the table is fully
        saved before continuing.
        """
        with self.fileLock:
            writer = threading.Thread(target=self._write,
                                      name="Preferences Write")
            writer.start()
            self.fileLock.wait()

    def _write(self):
        """Internal method that actually writes the table to a file. This is
        called in its own thread when :func:`save` is called.
        """
        with self.lock:
            with self.fileLock:
                self.fileLock.notify_all()

            with open(self.FILE_NAME, "w") as output:
                output.write("[Preferences]\n")
                for key in self.keylist:
                    value = self.values.get(key, "")
                    comment = self.comments.get(key, "")
                    if comment:
                        output.write(comment)

                    output.write(key)
                    output.write(self.VALUE_PREFIX)
                    output.write(value)
                    output.write(self.VALUE_SUFFIX)

                output.write(self.endComment)

            try:
                from networktables import NetworkTable
                NetworkTable.getTable(self.TABLE_NAME).putBoolean(self.SAVE_FIELD, False)
            except ImportError:
                pass

    def read(self):
        """The internal method to read from a file. This will be called in its
        own thread when the preferences singleton is first created.
        """

        with self.lock:
            with self.fileLock:
                self.fileLock.notify_all()

            comment = []

            try:
                with open(self.FILE_NAME) as f:
                    for line in f:
                        line = line.strip()
                        if not line:
                            comment.append(self.NEW_LINE)
                        elif line[0] == ';':
                            comment.append(line)
                            comment.append(self.NEW_LINE)
                        elif line[0] == '[':
                            continue # throw it away
                        else:
                            name, value = line.partition('=')
                            name = name.strip()
                            value = value.strip()

                            # Get between quotes if it starts with a quote
                            if value and value[0] == '"':
                                value = value[1:].partition('"')[0]

                            self.keylist.append(name)
                            self.values[name] = value
                            try:
                                from networktables import NetworkTable
                                NetworkTable.getTable(self.TABLE_NAME).putString(name, value)
                            except ImportError:
                                pass

                            if comment:
                                self.comments[name] = "".join(comment)
                                comment = []
            except FileNotFoundError:
                pass

            if comment:
                self.endComment = "".join(comment)

            logger.info("Done reading preferences")

        try:
            from networktables import NetworkTable
            NetworkTable.getTable(self.TABLE_NAME).putBoolean(self.SAVE_FIELD, False)
            # TODO: Verify that this works even though it changes with
            # subtables. Should work since preferences shouldn't have subtables.
            NetworkTable.getTable(self.TABLE_NAME).addTableListener(self.valueChanged)
        except ImportError:
            pass

    def valueChanged(self, source, key, value, isNew):
        if key == self.SAVE_FIELD:
            if value:
                self.save()
        else:
            with self.lock:
                if any((c in key) for c in "=\n\r\t[] ") or '"' in str(value):
                    if key in self.values or key in self.keylist:
                        self.values.pop(key, None)
                        try:
                            self.keylist.remove(key)
                        except ValueError:
                            pass
                        try:
                            from networktables import NetworkTable
                            NetworkTable.getTable(self.TABLE_NAME).putString(key, '"')
                        except ImportError:
                            pass
                else:
                    if key not in self.values:
                        self.keylist.append(key)
                    self.values[key] = str(value)



from .sendable import Sendable

__all__ = ["LiveWindowSendable"]

class LiveWindowSendable(Sendable):
    """A special type of object that can be displayed on the live window.
    """

    def updateTable(self):
        """Update the table for this sendable object with the latest
        values.
        """
        pass

    def startLiveWindowMode(self):
        """Start having this sendable object automatically respond to
        value changes reflect the value on the table.

        Default implementation will add self.valueChanged (if it exists)
        as a table listener on "Value".
        """
        if hasattr(self, "valueChanged"):
            table = getattr(self, "table", None)
            table_listener = getattr(self, "table_listener", None)
            if table is None or table_listener is not None:
                return
            self.table_listener = self.valueChanged
            table.addTableListener(self.valueChanged, True, key="Value")

    def stopLiveWindowMode(self):
        """Stop having this sendable object automatically respond to value
        changes.
        """
        # TODO: Broken, should only remove the listener from "Value" only.
        table = getattr(self, "table", None)
        table_listener = getattr(self, "table_listener", None)
        if table is None or table_listener is None:
            return
        table.removeTableListener(table_listener)
        self.table_listener = None



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

from .analoginput import AnalogInput
from .interfaces import PIDSource
from .livewindow import LiveWindow
from .sensorbase import SensorBase
from .timer import Timer

__all__ = ["Gyro"]

class Gyro(SensorBase):
    """Interface to a gyro device via an :class:`.AnalogInput`
    
    Use a rate gyro to return the robots heading relative to a starting
    position.  The Gyro class tracks the robots heading based on the starting
    position. As the robot rotates the new heading is computed by integrating
    the rate of rotation returned by the sensor. When the class is
    instantiated, it does a short calibration routine where it samples the
    gyro while at rest to determine the default offset. This is subtracted
    from each sample to determine the heading.

    .. not_implemented: initGyro
    """

    kOversampleBits = 10
    kAverageBits = 0
    kSamplesPerSecond = 50.0
    kCalibrationSampleTime = 5.0
    kDefaultVoltsPerDegreePerSecond = 0.007

    def __init__(self, channel):
        """Gyro constructor.

        Also initializes the gyro. Calibrate the gyro by running for a number
        of samples and computing the center value. Then use the
        center value as the Accumulator center value for subsequent
        measurements. It's important to make sure that the robot is not
        moving while the centering calculations are in progress, this is
        typically done when the robot is first turned on while it's sitting
        at rest before the competition starts.

        :param channel: The analog channel index or AnalogInput object that
            the gyro is connected to. Gyros can only be used on on-board channels 0-1.
        """
        if not hasattr(channel, "initAccumulator"):
            channel = AnalogInput(channel)
        self.analog = channel

        self.voltsPerDegreePerSecond = Gyro.kDefaultVoltsPerDegreePerSecond
        self.analog.setAverageBits(Gyro.kAverageBits)
        self.analog.setOversampleBits(Gyro.kOversampleBits)
        sampleRate = Gyro.kSamplesPerSecond \
                * (1 << (Gyro.kAverageBits + Gyro.kOversampleBits))
        AnalogInput.setGlobalSampleRate(sampleRate)
        Timer.delay(1.0)

        self.analog.initAccumulator()
        self.analog.resetAccumulator()
        
        # Only do this on a real robot
        if not hal.HALIsSimulation():
            Timer.delay(Gyro.kCalibrationSampleTime)

        value, count = self.analog.getAccumulatorOutput()

        self.center = int(float(value) / float(count) + .5)

        self.offset = (float(value) / float(count)) - self.center

        self.analog.setAccumulatorCenter(self.center)
        self.analog.resetAccumulator()

        self.setDeadband(0.0)

        self.pidSource = PIDSource.PIDSourceParameter.kAngle

        hal.HALReport(hal.HALUsageReporting.kResourceType_Gyro,
                      self.analog.getChannel())
        LiveWindow.addSensorChannel("Gyro", self.analog.getChannel(), self)

    def reset(self):
        """Reset the gyro. Resets the gyro to a heading of zero. This can be
        used if there is significant drift in the gyro and it needs to be
        recalibrated after it has been running.
        """
        if self.analog is None:
            return
        self.analog.resetAccumulator()

    def free(self):
        """Delete (free) the accumulator and the analog components used for the
        gyro.
        """
        if self.analog is not None:
            self.analog.free()
            self.analog = None

    def getAngle(self):
        """Return the actual angle in degrees that the robot is currently
        facing.

        The angle is based on the current accumulator value corrected by the
        oversampling rate, the gyro type and the A/D calibration values. The
        angle is continuous, that is it will continue from 360 to 361 degrees. This allows
        algorithms that wouldn't want to see a discontinuity in the gyro output
        as it sweeps past from 360 to 0 on the second time around.

        :returns: The current heading of the robot in degrees. This heading is
                based on integration of the returned rate from the gyro.
        :rtype: float
        """
        if self.analog is None:
            return 0.0
        value, count = self.analog.getAccumulatorOutput()

        value -= count * self.offset

        return (value
                * 1e-9
                * self.analog.getLSBWeight()
                * (1 << self.analog.getAverageBits())
                / (AnalogInput.getGlobalSampleRate() * self.voltsPerDegreePerSecond))

    def getRate(self):
        """Return the rate of rotation of the gyro

        The rate is based on the most recent reading of the gyro analog value

        :returns: the current rate in degrees per second
        :rtype: float
        """
        if self.analog is None:
            return 0.0
        else:
            return ((self.analog.getAverageValue() - (self.center + self.offset))
                    * 1e-9
                    * self.analog.getLSBWeight()
                    / ((1 << self.analog.getOversampleBits()) * self.voltsPerDegreePerSecond))

    def setSensitivity(self, voltsPerDegreePerSecond):
        """Set the gyro sensitivity. This takes the number of
        volts/degree/second sensitivity of the gyro and uses it in subsequent
        calculations to allow the code to work with multiple gyros. This value
        is typically found in the gyro datasheet.

        :param voltsPerDegreePerSecond:
            The sensitivity in Volts/degree/second
        :type  voltsPerDegreePerSecond: float
        """
        self.voltsPerDegreePerSecond = voltsPerDegreePerSecond

    def setDeadband(self, volts):
        """Set the size of the neutral zone.  Any voltage from the gyro less
        than this amount from the center is considered stationary.  Setting a
        deadband will decrease the amount of drift when the gyro isn't
        rotating, but will make it less accurate.

        :param volts: The size of the deadband in volts
        :type  volts: float
        """
        if self.analog is None:
            return
        deadband = int(volts * 1e9 / self.analog.getLSBWeight() *
                       (1 << self.analog.getOversampleBits()))
        self.analog.setAccumulatorDeadband(deadband)

    def setPIDSourceParameter(self, pidSource):
        """Set which parameter of the gyro you are using as a process
        control variable. The Gyro class supports the rate and angle
        parameters.

        :param pidSource: An enum to select the parameter.
        :type  pidSource: :class:`.PIDSource.PIDSourceParameter`
        """
        if pidSource not in (PIDSource.PIDSourceParameter.kRate,
                             PIDSource.PIDSourceParameter.kAngle):
            raise ValueError("Must be kRate or kAngle")
        self.pidSource = pidSource

    def pidGet(self):
        """Get the output of the gyro for use with PIDControllers

        :returns: the current angle according to the gyro
        :rtype: float
        """
        if self.pidSource == PIDSource.PIDSourceParameter.kRate:
            return self.getRate()
        elif self.pidSource == PIDSource.PIDSourceParameter.kAngle:
            return self.getAngle()
        else:
            return 0.0

    # Live Window code, only does anything if live window is activated.

    def getSmartDashboardType(self):
        return "Gyro"

    def updateTable(self):
        table = self.getTable()
        if table is not None:
            table.putNumber("Value", self.getAngle())

    def startLiveWindowMode(self):
        pass

    def stopLiveWindowMode(self):
        pass



__all__ = ["Sendable"]

class Sendable:
    """The base interface for objects that can be sent over the network
    through network tables"""

    def initTable(self, subtable):
        """Initializes a table for this sendable object.

        :param subtable: The table to put the values in.
        """
        self.table = subtable
        if hasattr(self, "updateTable"):
            self.updateTable()

    def getTable(self):
        """:returns: the table that is currently associated with the sendable"""
        return getattr(self, "table", None)

    def getSmartDashboardType(self):
        """
            :returns: the string representation of the named data type that
                will be used by the smart dashboard for this sendable
        """
        raise NotImplementedError



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

from .analoginput import AnalogInput
from .livewindow import LiveWindow
from .livewindowsendable import LiveWindowSendable

__all__ = ["AnalogAccelerometer"]

class AnalogAccelerometer(LiveWindowSendable):
    """Analog Accelerometer
    
    The accelerometer reads acceleration directly through the sensor. Many
    sensors have multiple axis and can be treated as multiple devices. Each
    is calibrated by finding the center value over a period of time.
    
    .. not_implemented: initAccelerometer
    """

    def __init__(self, channel):
        """Create a new instance of Accelerometer from either an existing
        AnalogChannel or from an analog channel port index.

        :param channel: port index or an already initialized :class:`.AnalogInput`
        """
        if not hasattr(channel, "getAverageVoltage"):
            channel = AnalogInput(channel)
        self.analogChannel = channel
        self.voltsPerG = 1.0
        self.zeroGVoltage = 2.5
        hal.HALReport(hal.HALUsageReporting.kResourceType_Accelerometer,
                      self.analogChannel.getChannel())
        LiveWindow.addSensorChannel("Accelerometer",
                                    self.analogChannel.getChannel(), self)

    def getAcceleration(self):
        """Return the acceleration in Gs.

        The acceleration is returned units of Gs.

        :returns: The current acceleration of the sensor in Gs.
        :rtype: float
        """
        return (self.analogChannel.getAverageVoltage() - self.zeroGVoltage) / self.voltsPerG

    def setSensitivity(self, sensitivity):
        """Set the accelerometer sensitivity.

        This sets the sensitivity of the accelerometer used for calculating
        the acceleration.  The sensitivity varies by accelerometer model.
        There are constants defined for various models.

        :param sensitivity: The sensitivity of accelerometer in Volts per G.
        :type  sensitivity: float
        """
        self.voltsPerG = sensitivity

    def setZero(self, zero):
        """Set the voltage that corresponds to 0 G.

        The zero G voltage varies by accelerometer model. There are constants
        defined for various models.

        :param zero: The zero G voltage.
        :type  zero: float
        """
        self.zeroGVoltage = zero

    def pidGet(self):
        """Get the Acceleration for the PID Source parent.

        :returns: The current acceleration in Gs.
        :rtype: float
        """
        return self.getAcceleration()

    def getSmartDashboardType(self):
        return "Accelerometer"

    # Live Window code, only does anything if live window is activated.

    def updateTable(self):
        table = self.getTable()
        if table is not None:
            table.putNumber("Value", self.getAcceleration())

    def startLiveWindowMode(self):
        # Don't have to do anything special when entering the LiveWindow.
        pass

    def stopLiveWindowMode(self):
        # Don't have to do anything special when exiting the LiveWindow.
        pass



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
import weakref

from .livewindow import LiveWindow
from .resource import Resource
from .sensorbase import SensorBase

__all__ = ["Relay"]

def _freeRelay(port):
    hal.setRelayForward(port, False)
    hal.setRelayReverse(port, False)
    hal.freeDIO(port)

class Relay(SensorBase):
    """Controls VEX Robotics Spike style relay outputs.
    
    Relays are intended to be connected to Spikes or similar relays. The relay
    channels controls a pair of pins that are either both off, one on, the
    other on, or both on. This translates into two Spike outputs at 0v, one at
    12v and one at 0v, one at 0v and the other at 12v, or two Spike outputs at
    12V. This allows off, full forward, or full reverse control of motors without
    variable speed. It also allows the two channels (forward and reverse) to
    be used independently for something that does not care about voltage
    polarity (like a solenoid).
    
    .. not_implemented: initRelay
    """

    class Value:
        """The state to drive a Relay to."""
        
        #: Off
        kOff = 0
        
        #: On for relays with defined direction
        kOn = 1
        
        #: Forward
        kForward = 2
        
        #: Reverse
        kReverse = 3

    class Direction:
        """The Direction(s) that a relay is configured to operate in."""
        
        #: Both directions are valid
        kBoth = 0
        
        #: Only forward is valid
        kForward = 1
        
        #: Only reverse is valid
        kReverse = 2

    relayChannels = Resource(SensorBase.kRelayChannels * 2)

    def __init__(self, channel, direction=None):
        """Relay constructor given a channel.

        Initially the relay is set to both lines at 0v.

        :param channel: The channel number for this relay (0-3)
        :type  channel: int
        :param direction: The direction that the Relay object will control.
            If not specified, defaults to allowing both directions.
        :type  direction: :class:`Relay.Direction`
        """
        if direction is None:
            direction = self.Direction.kBoth
        self.channel = channel
        self.direction = direction

        self._initRelay()

        LiveWindow.addActuatorChannel("Relay", self.channel, self)
        self.set(self.Value.kOff)

    def _initRelay(self):
        SensorBase.checkRelayChannel(self.channel)
        try:
            if (self.direction == self.Direction.kBoth or
                self.direction == self.Direction.kForward):
                Relay.relayChannels.allocate(self, self.channel * 2)
                hal.HALReport(hal.HALUsageReporting.kResourceType_Relay,
                              self.channel)
            if (self.direction == self.Direction.kBoth or
                self.direction == self.Direction.kReverse):
                Relay.relayChannels.allocate(self, self.channel * 2 + 1)
                hal.HALReport(hal.HALUsageReporting.kResourceType_Relay,
                              self.channel + 128)
        except IndexError as e:
            raise IndexError("Relay channel %d is already allocated" % self.channel) from e

        self._port = hal.initializeDigitalPort(hal.getPort(self.channel))
        self._port_finalizer = weakref.finalize(self, _freeRelay, self._port)

    @property
    def port(self):
        if not self._port_finalizer.alive:
            return None
        return self._port

    def free(self):
        if (self.direction == self.Direction.kBoth or
            self.direction == self.Direction.kForward):
            Relay.relayChannels.free(self.channel*2)
        if (self.direction == self.Direction.kBoth or
            self.direction == self.Direction.kReverse):
            Relay.relayChannels.free(self.channel*2 + 1)

        self._port_finalizer()

    def set(self, value):
        """Set the relay state.

        Valid values depend on which directions of the relay are controlled by
        the object.

        When set to kBothDirections, the relay can be set to any of the four
        states: 0v-0v, 12v-0v, 0v-12v, 12v-12v

        When set to kForwardOnly or kReverseOnly, you can specify the constant
        for the direction or you can simply specify kOff and kOn. Using only
        kOff and kOn is recommended.

        :param value: The state to set the relay.
        :type  value: :class:`Relay.Value`
        """
        if self.port is None:
            raise ValueError("operation on freed port")
        if value == self.Value.kOff:
            if (self.direction == self.Direction.kBoth or
                self.direction == self.Direction.kForward):
                hal.setRelayForward(self.port, False)
            if (self.direction == self.Direction.kBoth or
                self.direction == self.Direction.kReverse):
                hal.setRelayReverse(self.port, False)
        elif value == self.Value.kOn:
            if (self.direction == self.Direction.kBoth or
                self.direction == self.Direction.kForward):
                hal.setRelayForward(self.port, True)
            if (self.direction == self.Direction.kBoth or
                self.direction == self.Direction.kReverse):
                hal.setRelayReverse(self.port, True)
        elif value == self.Value.kForward:
            if self.direction == self.Direction.kReverse:
                raise ValueError("A relay configured for reverse cannot be set to forward")
            if (self.direction == self.Direction.kBoth or
                self.direction == self.Direction.kForward):
                hal.setRelayForward(self.port, True)
            if self.direction == self.Direction.kBoth:
                hal.setRelayReverse(self.port, False)
        elif value == self.Value.kReverse:
            if self.direction == self.Direction.kForward:
                raise ValueError("A relay configured for forward cannot be set to reverse")
            if self.direction == self.Direction.kBoth:
                hal.setRelayForward(self.port, False)
            if (self.direction == self.Direction.kBoth or
                self.direction == self.Direction.kReverse):
                hal.setRelayReverse(self.port, True)
        else:
            raise ValueError("Invalid value argument '%s'" % value)

    def get(self):
        """Get the Relay State

        Gets the current state of the relay.

        When set to kForwardOnly or kReverseOnly, value is returned as kOn/kOff
        not kForward/kReverse (per the recommendation in Set)

        :returns: The current state of the relay
        :rtype: :class:`Relay.Value`
        """
        if self.port is None:
            raise ValueError("operation on freed port")
        if hal.getRelayForward(self.port):
            if hal.getRelayReverse(self.port):
                return self.Value.kOn
            else:
                if self.direction == self.Direction.kForward:
                    return self.Value.kOn
                else:
                    return self.Value.kForward
        else:
            if hal.getRelayReverse(self.port):
                if self.direction == self.Direction.kReverse:
                    return self.Value.kOn
                else:
                    return self.Value.kReverse
            else:
                return self.Value.kOff

    def setDirection(self, direction):
        """Set the Relay Direction.

        Changes which values the relay can be set to depending on which
        direction is used.

        Valid inputs are kBothDirections, kForwardOnly, and kReverseOnly.

        :param direction: The direction for the relay to operate in
        :type  direction: :class:`Relay.Direction`
        """
        if self.direction == direction:
            return
        
        if direction not in [self.Direction.kBoth,
                             self.Direction.kForward,
                             self.Direction.kReverse]:
            raise ValueError("Invalid direction argument '%s'" % direction)

        self.free()
        self.direction = direction
        self._initRelay()

    # Live Window code, only does anything if live window is activated.
    def getSmartDashboardType(self):
        return "Relay"

    def updateTable(self):
        table = self.getTable()
        if table is None:
            return
        v = self.get()
        if v == self.Value.kOn:
            table.putString("Value", "On")
        elif v == self.Value.kForward:
            table.putString("Value", "Forward")
        elif v == self.Value.kReverse:
            table.putString("Value", "Reverse")
        else:
            table.putString("Value", "Off")

    def valueChanged(self, itable, key, value, bln):
        if value == "Off":
            self.set(self.Value.kOff)
        elif value == "On":
            self.set(self.Value.kOn)
        elif value == "Forward":
            self.set(self.Value.kForward)
        elif value == "Reverse":
            self.set(self.Value.kReverse)



# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.

import hal
import warnings

import logging
logger = logging.getLogger('robotpy')

__all__ = ["RobotBase"]

class RobotBase:
    """Implement a Robot Program framework.
    
    The RobotBase class is intended to be subclassed by a user creating a
    robot program.  Overridden autonomous() and operatorControl() methods
    are called at the appropriate time as the match proceeds. In the current
    implementation, the Autonomous code will run to completion before the
    OperatorControl code could start. In the future the Autonomous code might
    be spawned as a task, then killed at the end of the Autonomous period.
    
    User code should be placed in the constructor that runs before the
    Autonomous or Operator Control period starts. The constructor will
    run to completion before Autonomous is entered.

    .. warning:: If you override __init__ in your robot class, you must call
                 the base class constructor. This must be used to ensure that
                 the communications code starts.
    """

    def __init__(self):
        # TODO: StartCAPI()
        # TODO: See if the next line is necessary
        # Resource.RestartProgram()

        try:
            from networktables import NetworkTable
            #NetworkTable.setServerMode() -- don't set this explicitly, it's the default.
        except ImportError:
            warnings.warn("networktables not found", ImportWarning)
            NetworkTable = None

        from .driverstation import DriverStation
        self.ds = DriverStation.getInstance()

        if NetworkTable is not None:
            NetworkTable.getTable("")   # forces network tables to initialize
            NetworkTable.getTable("LiveWindow").getSubTable("~STATUS~").putBoolean("LW Enabled", False)

        self.__initialized = True

    def prestart(self):
        """
        This hook is called right before :meth:`startCompetition`. By default, tell
        the DS that the robot is now ready to be enabled. If you don't want the
        robot to be enabled yet, you can override this method to do nothing.
        If you do so, you will need to call hal.HALNetworkCommunicationObserveUserProgramStarting()
        from your code when you are ready for the robot to be enabled.
        """
        hal.HALNetworkCommunicationObserveUserProgramStarting()


    def free(self):
        """Free the resources for a RobotBase class."""
        # TODO: delete?
        pass

    @staticmethod
    def isSimulation():
        """
            :returns: If the robot is running in simulation.
            :rtype: bool
        """
        return hal.HALIsSimulation()

    @staticmethod
    def isReal():
        """
            :returns: If the robot is running in the real world.
            :rtype: bool
        """
        return not hal.HALIsSimulation()

    def isDisabled(self):
        """Determine if the Robot is currently disabled.

        :returns: True if the Robot is currently disabled by the field
            controls.
        :rtype: bool
        """
        return self.ds.isDisabled()

    def isEnabled(self):
        """Determine if the Robot is currently enabled.

        :returns: True if the Robot is currently enabled by the field
            controls.
        :rtype: bool
        """
        return self.ds.isEnabled()

    def isAutonomous(self):
        """Determine if the robot is currently in Autonomous mode.

        :returns: True if the robot is currently operating Autonomously as
            determined by the field controls.
        :rtype: bool
        """
        return self.ds.isAutonomous()

    def isTest(self):
        """Determine if the robot is currently in Test mode.

        :returns: True if the robot is currently operating in Test mode as
            determined by the driver station.
        :rtype: bool
        """
        return self.ds.isTest()

    def isOperatorControl(self):
        """Determine if the robot is currently in Operator Control mode.

        :returns: True if the robot is currently operating in Tele-Op mode as
            determined by the field controls.
        :rtype: bool
        """
        return self.ds.isOperatorControl()

    def isNewDataAvailable(self):
        """Indicates if new data is available from the driver station.

        :returns: Has new data arrived over the network since the last time
            this function was called?
        :rtype: bool
        """
        return self.ds.isNewControlData()

    def startCompetition(self):
        """Provide an alternate "main loop" via startCompetition()."""
        raise NotImplementedError

    @staticmethod
    def initializeHardwareConfiguration():
        """Common initialization for all robot programs."""
        hal.HALInitialize()

        from .driverstation import DriverStation
        from .robotstate import RobotState
        RobotState.impl = DriverStation.getInstance()

    @staticmethod
    def main(robot_cls):
        """Starting point for the applications."""
        RobotBase.initializeHardwareConfiguration()

        hal.HALReport(hal.HALUsageReporting.kResourceType_Language,
                      hal.HALUsageReporting.kLanguage_Python)

        try:
            robot = robot_cls()
            robot.prestart()
        except:
            from .driverstation import DriverStation
            DriverStation.reportError("ERROR Could not instantiate robot", True)
            logger.exception("Robots don't quit!")
            logger.exception("Could not instantiate robot "+robot_cls.__name__+"!")
            return False
        
        # Add a check to see if the user forgot to call super().__init__()
        if not hasattr(robot, '_RobotBase__initialized'):
            logger.error("If your robot class has an __init__ function, it must call super().__init__()!")
            return False
        
        if not hal.HALIsSimulation():
            try:
                import wpilib
                with open('/tmp/frc_versions/FRC_Lib_Version.ini', 'w') as fp:
                    fp.write('RobotPy %s' % wpilib.__version__)
            except:
                logger.warning("Could not write FRC version file to disk")

        try:
            robot.startCompetition()
        except KeyboardInterrupt:
            logger.exception("THIS IS NOT AN ERROR: The user hit CTRL-C to kill the robot")
            logger.info("Exiting because of keyboard interrupt")
            return True
        except:
            from .driverstation import DriverStation
            DriverStation.reportError("ERROR Unhandled exception", True)
            logger.warn("Robots don't quit!")
            logger.exception("---> The startCompetition() method (or methods called by it) should have handled the exception.")
            return False
        else:
            # startCompetition never returns unless exception occurs....
            from .driverstation import DriverStation
            DriverStation.reportError("ERROR startCompetition() returned", False)
            logger.warn("Robots don't quit!")
            logger.error("---> Unexpected return from startCompetition() method.")
            return False


#----------------------------------------------------------------------------*/
# Copyright (c) FIRST 2008-2014. All Rights Reserved.                        */
# Open Source Software - may be modified and shared by FRC teams. The code   */
# must be accompanied by the FIRST BSD license file in the root directory of */
# the project.                                                               */
#----------------------------------------------------------------------------*/

import hal

from .analoginput import AnalogInput
from .livewindowsendable import LiveWindowSendable

__all__ = ["AnalogPotentiometer"]

class AnalogPotentiometer(LiveWindowSendable):
    """Reads a potentiometer via an :class:`.AnalogInput`
    
    Analog potentiometers read
    in an analog voltage that corresponds to a position. The position is in
    whichever units you choose, by way of the scaling and offset constants
    passed to the constructor.

    .. not_implemented: initPot
    """

    def __init__(self, channel, fullRange=1.0, offset=0.0):
        """AnalogPotentiometer constructor.

        Use the fullRange and offset values so that the output produces
        meaningful values. I.E: you have a 270 degree potentiometer and
        you want the output to be degrees with the halfway point as 0
        degrees. The fullRange value is 270.0(degrees) and the offset is
        -135.0 since the halfway point after scaling is 135 degrees.

        :param channel: The analog channel this potentiometer is plugged into.
        :type  channel: int or :class:`.AnalogInput`
        :param fullRange: The scaling to multiply the fraction by to get a
            meaningful unit.  Defaults to 1.0 if unspecified.
        :type  fullRange: float
        :param offset: The offset to add to the scaled value for controlling
            the zero value.  Defaults to 0.0 if unspecified.
        :type  offset: float
        """

        if not hasattr(channel, "getVoltage"):
            channel = AnalogInput(channel)
        self.analog_input = channel
        self.fullRange = fullRange
        self.offset = offset
        self.init_analog_input = True

    def get(self):
        """Get the current reading of the potentiometer.

        :returns: The current position of the potentiometer.
        :rtype: float
        """
        return (self.analog_input.getVoltage() / hal.getUserVoltage5V()) * self.fullRange + self.offset

    def pidGet(self):
        """Implement the PIDSource interface.

        :returns: The current reading.
        :rtype: float
        """
        return self.get()

    # Live Window code, only does anything if live window is activated.

    def getSmartDashboardType(self):
        return "Analog Input"

    def updateTable(self):
        table = self.getTable()
        if table is not None:
            table.putNumber("Value", self.get())

    def startLiveWindowMode(self):
        # don't have to do anything special when entering the LiveWindow
        pass

    def stopLiveWindowMode(self):
        # don't have to do anything special when exiting the LiveWindow
        pass

    def free(self):
        if self.init_analog_input:
            self.analog_input.free()
            del self.analog_input
            self.init_analog_input = False



# The masks of the fields that are used in the message identifier.
CAN_MSGID_FULL_M        = 0x1fffffff
CAN_MSGID_DEVNO_M       = 0x0000003f
CAN_MSGID_API_M         = 0x0000ffc0
CAN_MSGID_MFR_M         = 0x00ff0000
CAN_MSGID_DTYPE_M       = 0x1f000000
CAN_MSGID_DEVNO_S       = 0
CAN_MSGID_API_S         = 6
CAN_MSGID_MFR_S         = 16
CAN_MSGID_DTYPE_S       = 24

# The Reserved device number values in the Message Id.
CAN_MSGID_DEVNO_BCAST   = 0x00000000

# The Reserved system control API numbers in the Message Id.
CAN_MSGID_API_SYSHALT   = 0x00000000
CAN_MSGID_API_SYSRST    = 0x00000040
CAN_MSGID_API_DEVASSIGN = 0x00000080
CAN_MSGID_API_DEVQUERY  = 0x000000c0
CAN_MSGID_API_HEARTBEAT = 0x00000140
CAN_MSGID_API_SYNC      = 0x00000180
CAN_MSGID_API_UPDATE    = 0x000001c0
CAN_MSGID_API_FIRMVER   = 0x00000200
CAN_MSGID_API_ENUMERATE = 0x00000240
CAN_MSGID_API_SYSRESUME = 0x00000280

# The 32 bit values associated with the CAN_MSGID_API_STATUS request.
CAN_STATUS_CODE_M       = 0x0000ffff
CAN_STATUS_MFG_M        = 0x00ff0000
CAN_STATUS_DTYPE_M      = 0x1f000000
CAN_STATUS_CODE_S       = 0
CAN_STATUS_MFG_S        = 16
CAN_STATUS_DTYPE_S      = 24

# The Reserved manufacturer identifiers in the Message Id.
CAN_MSGID_MFR_NI        = 0x00010000
CAN_MSGID_MFR_LM        = 0x00020000
CAN_MSGID_MFR_DEKA      = 0x00030000

# The Reserved device type identifiers in the Message Id.
CAN_MSGID_DTYPE_BCAST   = 0x00000000
CAN_MSGID_DTYPE_ROBOT   = 0x01000000
CAN_MSGID_DTYPE_MOTOR   = 0x02000000
CAN_MSGID_DTYPE_RELAY   = 0x03000000
CAN_MSGID_DTYPE_GYRO    = 0x04000000
CAN_MSGID_DTYPE_ACCEL   = 0x05000000
CAN_MSGID_DTYPE_USONIC  = 0x06000000
CAN_MSGID_DTYPE_GEART   = 0x07000000
CAN_MSGID_DTYPE_UPDATE  = 0x1f000000

# LM Motor Control API Classes API Class and ID masks.
CAN_MSGID_API_CLASS_M   = 0x0000fc00
CAN_MSGID_API_ID_M      = 0x000003c0

# LM Motor Control API Classes in the Message Id for non-broadcast.
# These are the upper 6 bits of the API field, the lower 4 bits determine
# the APIId.
CAN_API_MC_VOLTAGE      = 0x00000000
CAN_API_MC_SPD          = 0x00000400
CAN_API_MC_VCOMP        = 0x00000800
CAN_API_MC_POS          = 0x00000c00
CAN_API_MC_ICTRL        = 0x00001000
CAN_API_MC_STATUS       = 0x00001400
CAN_API_MC_PSTAT        = 0x00001800
CAN_API_MC_CFG          = 0x00001c00
CAN_API_MC_ACK          = 0x00002000

# The Stellaris Motor Class Control Voltage API definitions.
LM_API_VOLT             = (CAN_MSGID_MFR_LM | CAN_MSGID_DTYPE_MOTOR |   \
                             CAN_API_MC_VOLTAGE)
LM_API_VOLT_EN          = (LM_API_VOLT | (0 << CAN_MSGID_API_S))
LM_API_VOLT_DIS         = (LM_API_VOLT | (1 << CAN_MSGID_API_S))
LM_API_VOLT_SET         = (LM_API_VOLT | (2 << CAN_MSGID_API_S))
LM_API_VOLT_SET_RAMP    = (LM_API_VOLT | (3 << CAN_MSGID_API_S))
###### FIRST BEGIN #####
LM_API_VOLT_T_EN        = (LM_API_VOLT | (4 << CAN_MSGID_API_S))
LM_API_VOLT_T_SET       = (LM_API_VOLT | (5 << CAN_MSGID_API_S))
LM_API_VOLT_T_SET_NO_ACK = (LM_API_VOLT | (7 << CAN_MSGID_API_S))
###### FIRST END #####
LM_API_VOLT_SET_NO_ACK  = (LM_API_VOLT | (8 << CAN_MSGID_API_S))

# The Stellaris Motor Class Control API definitions for LM_API_VOLT_SET_RAMP.
LM_API_VOLT_RAMP_DIS    = 0

# The Stellaris Motor Class Control API definitions for CAN_MSGID_API_SYNC.
LM_API_SYNC_PEND_NOW    = 0

# The Stellaris Motor Class Speed Control API definitions.
LM_API_SPD              = (CAN_MSGID_MFR_LM | CAN_MSGID_DTYPE_MOTOR |   \
                             CAN_API_MC_SPD)
LM_API_SPD_EN           = (LM_API_SPD | (0 << CAN_MSGID_API_S))
LM_API_SPD_DIS          = (LM_API_SPD | (1 << CAN_MSGID_API_S))
LM_API_SPD_SET          = (LM_API_SPD | (2 << CAN_MSGID_API_S))
LM_API_SPD_PC           = (LM_API_SPD | (3 << CAN_MSGID_API_S))
LM_API_SPD_IC           = (LM_API_SPD | (4 << CAN_MSGID_API_S))
LM_API_SPD_DC           = (LM_API_SPD | (5 << CAN_MSGID_API_S))
LM_API_SPD_REF          = (LM_API_SPD | (6 << CAN_MSGID_API_S))
###### FIRST BEGIN #####
LM_API_SPD_T_EN         = (LM_API_SPD | (7 << CAN_MSGID_API_S))
LM_API_SPD_T_SET        = (LM_API_SPD | (8 << CAN_MSGID_API_S))
LM_API_SPD_T_SET_NO_ACK = (LM_API_SPD | (10 << CAN_MSGID_API_S))
###### FIRST END #####
LM_API_SPD_SET_NO_ACK   = (LM_API_SPD | (11 << CAN_MSGID_API_S))

# The Stellaris Motor Control Voltage Compensation Control API definitions.
LM_API_VCOMP            = (CAN_MSGID_MFR_LM | CAN_MSGID_DTYPE_MOTOR |   \
                             CAN_API_MC_VCOMP)
LM_API_VCOMP_EN         = (LM_API_VCOMP | (0 << CAN_MSGID_API_S))
LM_API_VCOMP_DIS        = (LM_API_VCOMP | (1 << CAN_MSGID_API_S))
LM_API_VCOMP_SET        = (LM_API_VCOMP | (2 << CAN_MSGID_API_S))
LM_API_VCOMP_IN_RAMP    = (LM_API_VCOMP | (3 << CAN_MSGID_API_S))
LM_API_VCOMP_COMP_RAMP  = (LM_API_VCOMP | (4 << CAN_MSGID_API_S))
###### FIRST BEGIN #####
LM_API_VCOMP_T_EN       = (LM_API_VCOMP | (5 << CAN_MSGID_API_S))
LM_API_VCOMP_T_SET      = (LM_API_VCOMP | (6 << CAN_MSGID_API_S))
LM_API_VCOMP_T_SET_NO_ACK = (LM_API_VCOMP | (8 << CAN_MSGID_API_S))
###### FIRST END #####
LM_API_VCOMP_SET_NO_ACK = (LM_API_VCOMP | (9 << CAN_MSGID_API_S))

# The Stellaris Motor Class Position Control API definitions.
LM_API_POS              = (CAN_MSGID_MFR_LM | CAN_MSGID_DTYPE_MOTOR |   \
                             CAN_API_MC_POS)
LM_API_POS_EN           = (LM_API_POS | (0 << CAN_MSGID_API_S))
LM_API_POS_DIS          = (LM_API_POS | (1 << CAN_MSGID_API_S))
LM_API_POS_SET          = (LM_API_POS | (2 << CAN_MSGID_API_S))
LM_API_POS_PC           = (LM_API_POS | (3 << CAN_MSGID_API_S))
LM_API_POS_IC           = (LM_API_POS | (4 << CAN_MSGID_API_S))
LM_API_POS_DC           = (LM_API_POS | (5 << CAN_MSGID_API_S))
LM_API_POS_REF          = (LM_API_POS | (6 << CAN_MSGID_API_S))
###### FIRST BEGIN #####
LM_API_POS_T_EN         = (LM_API_POS | (7 << CAN_MSGID_API_S))
LM_API_POS_T_SET        = (LM_API_POS | (8 << CAN_MSGID_API_S))
LM_API_POS_T_SET_NO_ACK = (LM_API_POS | (10 << CAN_MSGID_API_S))
###### FIRST END #####
LM_API_POS_SET_NO_ACK   = (LM_API_POS | (11 << CAN_MSGID_API_S))

# The Stellaris Motor Class Current Control API definitions.
LM_API_ICTRL            = (CAN_MSGID_MFR_LM | CAN_MSGID_DTYPE_MOTOR |   \
                             CAN_API_MC_ICTRL)
LM_API_ICTRL_EN         = (LM_API_ICTRL | (0 << CAN_MSGID_API_S))
LM_API_ICTRL_DIS        = (LM_API_ICTRL | (1 << CAN_MSGID_API_S))
LM_API_ICTRL_SET        = (LM_API_ICTRL | (2 << CAN_MSGID_API_S))
LM_API_ICTRL_PC         = (LM_API_ICTRL | (3 << CAN_MSGID_API_S))
LM_API_ICTRL_IC         = (LM_API_ICTRL | (4 << CAN_MSGID_API_S))
LM_API_ICTRL_DC         = (LM_API_ICTRL | (5 << CAN_MSGID_API_S))
###### FIRST BEGIN #####
LM_API_ICTRL_T_EN       = (LM_API_ICTRL | (6 << CAN_MSGID_API_S))
LM_API_ICTRL_T_SET      = (LM_API_ICTRL | (7 << CAN_MSGID_API_S))
LM_API_ICTRL_T_SET_NO_ACK = (LM_API_ICTRL | (9 << CAN_MSGID_API_S))
###### FIRST END #####
LM_API_ICTRL_SET_NO_ACK = (LM_API_ICTRL | (10 << CAN_MSGID_API_S))

# The Stellaris Motor Class Firmware Update API definitions.
LM_API_UPD              = (CAN_MSGID_MFR_LM | CAN_MSGID_DTYPE_UPDATE)
LM_API_UPD_PING         = (LM_API_UPD | (0 << CAN_MSGID_API_S))
LM_API_UPD_DOWNLOAD     = (LM_API_UPD | (1 << CAN_MSGID_API_S))
LM_API_UPD_SEND_DATA    = (LM_API_UPD | (2 << CAN_MSGID_API_S))
LM_API_UPD_RESET        = (LM_API_UPD | (3 << CAN_MSGID_API_S))
LM_API_UPD_ACK          = (LM_API_UPD | (4 << CAN_MSGID_API_S))
LM_API_HWVER            = (LM_API_UPD | (5 << CAN_MSGID_API_S))
LM_API_UPD_REQUEST      = (LM_API_UPD | (6 << CAN_MSGID_API_S))
###### FIRST BEGIN #####
LM_API_UNTRUST_EN       = (LM_API_UPD | (11 << CAN_MSGID_API_S))
LM_API_TRUST_EN         = (LM_API_UPD | (12 << CAN_MSGID_API_S))
LM_API_TRUST_HEARTBEAT  = (LM_API_UPD | (13 << CAN_MSGID_API_S))
###### FIRST END #####

# The Stellaris Motor Class Status API definitions.
LM_API_STATUS           = (CAN_MSGID_MFR_LM | CAN_MSGID_DTYPE_MOTOR |   \
                             CAN_API_MC_STATUS)
LM_API_STATUS_VOLTOUT   = (LM_API_STATUS | (0 << CAN_MSGID_API_S))
LM_API_STATUS_VOLTBUS   = (LM_API_STATUS | (1 << CAN_MSGID_API_S))
LM_API_STATUS_CURRENT   = (LM_API_STATUS | (2 << CAN_MSGID_API_S))
LM_API_STATUS_TEMP      = (LM_API_STATUS | (3 << CAN_MSGID_API_S))
LM_API_STATUS_POS       = (LM_API_STATUS | (4 << CAN_MSGID_API_S))
LM_API_STATUS_SPD       = (LM_API_STATUS | (5 << CAN_MSGID_API_S))
LM_API_STATUS_LIMIT     = (LM_API_STATUS | (6 << CAN_MSGID_API_S))
LM_API_STATUS_FAULT     = (LM_API_STATUS | (7 << CAN_MSGID_API_S))
LM_API_STATUS_POWER     = (LM_API_STATUS | (8 << CAN_MSGID_API_S))
LM_API_STATUS_CMODE     = (LM_API_STATUS | (9 << CAN_MSGID_API_S))
LM_API_STATUS_VOUT      = (LM_API_STATUS | (10 << CAN_MSGID_API_S))
LM_API_STATUS_STKY_FLT  = (LM_API_STATUS | (11 << CAN_MSGID_API_S))
LM_API_STATUS_FLT_COUNT = (LM_API_STATUS | (12 << CAN_MSGID_API_S))

# These definitions are used with the byte that is returned from
# the status request for LM_API_STATUS_LIMIT.
LM_STATUS_LIMIT_FWD         = 0x01
LM_STATUS_LIMIT_REV         = 0x02
LM_STATUS_LIMIT_SFWD        = 0x04
LM_STATUS_LIMIT_SREV        = 0x08
LM_STATUS_LIMIT_STKY_FWD    = 0x10
LM_STATUS_LIMIT_STKY_REV    = 0x20
LM_STATUS_LIMIT_STKY_SFWD   = 0x40
LM_STATUS_LIMIT_STKY_SREV   = 0x80

# LM Motor Control status codes returned due to the CAN_STATUS_CODE_M field.
LM_STATUS_FAULT_ILIMIT  = 0x01
LM_STATUS_FAULT_TLIMIT  = 0x02
LM_STATUS_FAULT_VLIMIT  = 0x04

# The Stellaris Motor Class Configuration API definitions.
LM_API_CFG              = (CAN_MSGID_MFR_LM | CAN_MSGID_DTYPE_MOTOR |   \
                             CAN_API_MC_CFG)
LM_API_CFG_NUM_BRUSHES  = (LM_API_CFG | (0 << CAN_MSGID_API_S))
LM_API_CFG_ENC_LINES    = (LM_API_CFG | (1 << CAN_MSGID_API_S))
LM_API_CFG_POT_TURNS    = (LM_API_CFG | (2 << CAN_MSGID_API_S))
LM_API_CFG_BRAKE_COAST  = (LM_API_CFG | (3 << CAN_MSGID_API_S))
LM_API_CFG_LIMIT_MODE   = (LM_API_CFG | (4 << CAN_MSGID_API_S))
LM_API_CFG_LIMIT_FWD    = (LM_API_CFG | (5 << CAN_MSGID_API_S))
LM_API_CFG_LIMIT_REV    = (LM_API_CFG | (6 << CAN_MSGID_API_S))
LM_API_CFG_MAX_VOUT     = (LM_API_CFG | (7 << CAN_MSGID_API_S))
LM_API_CFG_FAULT_TIME   = (LM_API_CFG | (8 << CAN_MSGID_API_S))

# The Stellaris ACK API definition.
LM_API_ACK              = (CAN_MSGID_MFR_LM | CAN_MSGID_DTYPE_MOTOR |   \
                             CAN_API_MC_ACK)

# The 8 bit values that can be returned by a call to LM_API_STATUS_HWVER.
LM_HWVER_UNKNOWN        = 0x00
LM_HWVER_JAG_1_0        = 0x01
LM_HWVER_JAG_2_0        = 0x02

# The 8 bit values that can be returned by a call to LM_API_STATUS_CMODE.
LM_STATUS_CMODE_VOLT    = 0x00
LM_STATUS_CMODE_CURRENT = 0x01
LM_STATUS_CMODE_SPEED   = 0x02
LM_STATUS_CMODE_POS     = 0x03
LM_STATUS_CMODE_VCOMP   = 0x04

# The values that can specified as the position or speed reference.  Not all
# values are valid for each reference; if an invalid reference is set, then
# none will be selected.
LM_REF_ENCODER          = 0x00
LM_REF_POT              = 0x01
LM_REF_INV_ENCODER      = 0x02
LM_REF_QUAD_ENCODER     = 0x03
LM_REF_NONE             = 0xff

# The flags that are used to indicate the currently active fault sources.
LM_FAULT_CURRENT        = 0x01
LM_FAULT_TEMP           = 0x02
LM_FAULT_VBUS           = 0x04
LM_FAULT_GATE_DRIVE     = 0x08
LM_FAULT_COMM           = 0x10

# The Stellaris Motor Class Periodic Status API definitions.
LM_API_PSTAT            = (CAN_MSGID_MFR_LM | CAN_MSGID_DTYPE_MOTOR |   \
                             CAN_API_MC_PSTAT)
LM_API_PSTAT_PER_EN_S0  = (LM_API_PSTAT | (0 << CAN_MSGID_API_S))
LM_API_PSTAT_PER_EN_S1  = (LM_API_PSTAT | (1 << CAN_MSGID_API_S))
LM_API_PSTAT_PER_EN_S2  = (LM_API_PSTAT | (2 << CAN_MSGID_API_S))
LM_API_PSTAT_PER_EN_S3  = (LM_API_PSTAT | (3 << CAN_MSGID_API_S))
LM_API_PSTAT_CFG_S0     = (LM_API_PSTAT | (4 << CAN_MSGID_API_S))
LM_API_PSTAT_CFG_S1     = (LM_API_PSTAT | (5 << CAN_MSGID_API_S))
LM_API_PSTAT_CFG_S2     = (LM_API_PSTAT | (6 << CAN_MSGID_API_S))
LM_API_PSTAT_CFG_S3     = (LM_API_PSTAT | (7 << CAN_MSGID_API_S))
LM_API_PSTAT_DATA_S0    = (LM_API_PSTAT | (8 << CAN_MSGID_API_S))
LM_API_PSTAT_DATA_S1    = (LM_API_PSTAT | (9 << CAN_MSGID_API_S))
LM_API_PSTAT_DATA_S2    = (LM_API_PSTAT | (10 << CAN_MSGID_API_S))
LM_API_PSTAT_DATA_S3    = (LM_API_PSTAT | (11 << CAN_MSGID_API_S))

# The values that can be used to configure the data the Periodic Status
# Message bytes.  Bytes of a multi-byte data values are encoded as
# little-endian, therefore B0 is the least significant byte.
LM_PSTAT_END            = 0
LM_PSTAT_VOLTOUT_B0     = 1
LM_PSTAT_VOLTOUT_B1     = 2
LM_PSTAT_VOLTBUS_B0     = 3
LM_PSTAT_VOLTBUS_B1     = 4
LM_PSTAT_CURRENT_B0     = 5
LM_PSTAT_CURRENT_B1     = 6
LM_PSTAT_TEMP_B0        = 7
LM_PSTAT_TEMP_B1        = 8
LM_PSTAT_POS_B0         = 9
LM_PSTAT_POS_B1         = 10
LM_PSTAT_POS_B2         = 11
LM_PSTAT_POS_B3         = 12
LM_PSTAT_SPD_B0         = 13
LM_PSTAT_SPD_B1         = 14
LM_PSTAT_SPD_B2         = 15
LM_PSTAT_SPD_B3         = 16
LM_PSTAT_LIMIT_NCLR     = 17
LM_PSTAT_LIMIT_CLR      = 18
LM_PSTAT_FAULT          = 19
LM_PSTAT_STKY_FLT_NCLR  = 20
LM_PSTAT_STKY_FLT_CLR   = 21
LM_PSTAT_VOUT_B0        = 22
LM_PSTAT_VOUT_B1        = 23
LM_PSTAT_FLT_COUNT_CURRENT = 24
LM_PSTAT_FLT_COUNT_TEMP = 25
LM_PSTAT_FLT_COUNT_VOLTBUS = 26
LM_PSTAT_FLT_COUNT_GATE = 27
LM_PSTAT_FLT_COUNT_COMM = 28
LM_PSTAT_CANSTS         = 29
LM_PSTAT_CANERR_B0      = 30
LM_PSTAT_CANERR_B1      = 31



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal
import weakref

from .analogtriggeroutput import AnalogTriggerOutput
from .resource import Resource

__all__ = ["AnalogTrigger"]

def _freeAnalogTrigger(port):
    hal.cleanAnalogTrigger(port)

class AnalogTrigger:
    """
        Converts an analog signal into a digital signal
        
        An analog trigger is a way to convert an analog signal into a digital
        signal using resources built into the FPGA. The resulting digital
        signal can then be used directly or fed into other digital components
        of the FPGA such as the counter or encoder modules. The analog trigger
        module works by comparing analog signals to a voltage range set by
        the code. The specific return types and meanings depend on the analog
        trigger mode in use.
        
        .. not_implemented: initTrigger
    """

    AnalogTriggerType = AnalogTriggerOutput.AnalogTriggerType

    def __init__(self, channel):
        """Constructor for an analog trigger given a channel number or analog
        input.

        :param channel: the port index or :class:`.AnalogInput` to use for the analog
            trigger.  Treated as an AnalogInput if the provided object has a
            getChannel function.
        """
        if hasattr(channel, "getChannel"):
            channel = channel.getChannel()

        port = hal.getPort(channel)
        self._port, self.index = hal.initializeAnalogTrigger(port)
        self._analogtrigger_finalizer = \
                weakref.finalize(self, _freeAnalogTrigger, self._port)
                
        # Need this to free on unit test wpilib reset
        Resource._add_global_resource(self)

        hal.HALReport(hal.HALUsageReporting.kResourceType_AnalogTrigger,
                      channel)

    @property
    def port(self):
        if not self._analogtrigger_finalizer.alive:
            return None
        return self._port

    def free(self):
        """Release the resources used by this object"""
        self._analogtrigger_finalizer()

    def setLimitsRaw(self, lower, upper):
        """Set the upper and lower limits of the analog trigger. The limits are
        given in ADC codes. If oversampling is used, the units must be scaled
        appropriately.

        :param lower: the lower raw limit
        :param upper: the upper raw limit
        """
        if lower > upper:
            raise ValueError("Lower bound is greater than upper")
        if self.port is None:
            raise ValueError("operation on freed port")
        hal.setAnalogTriggerLimitsRaw(self.port, lower, upper)

    def setLimitsVoltage(self, lower, upper):
        """Set the upper and lower limits of the analog trigger. The limits are
        given as floating point voltage values.

        :param lower: the lower voltage limit
        :param upper: the upper voltage limit
        """
        if lower > upper:
            raise ValueError("Lower bound is greater than upper")
        if self.port is None:
            raise ValueError("operation on freed port")
        hal.setAnalogTriggerLimitsVoltage(self.port, float(lower), float(upper))

    def setAveraged(self, useAveragedValue):
        """Configure the analog trigger to use the averaged vs. raw values. If
        the value is true, then the averaged value is selected for the analog
        trigger, otherwise the immediate value is used.

        :param useAveragedValue: True to use an averaged value, False otherwise
        """
        if self.port is None:
            raise ValueError("operation on freed port")
        hal.setAnalogTriggerAveraged(self.port, useAveragedValue)

    def setFiltered(self, useFilteredValue):
        """Configure the analog trigger to use a filtered value. The analog
        trigger will operate with a 3 point average rejection filter. This is
        designed to help with 360 degree pot applications for the period where
        the pot crosses through zero.

        :param useFilteredValue: True to use a filterd value, False otherwise
        """
        if self.port is None:
            raise ValueError("operation on freed port")
        hal.setAnalogTriggerFiltered(self.port, useFilteredValue)

    def getIndex(self):
        """Return the index of the analog trigger. This is the FPGA index of
        this analog trigger instance.

        :returns: The index of the analog trigger.
        """
        if self.port is None:
            raise ValueError("operation on freed port")
        return self.index

    def getInWindow(self):
        """Return the InWindow output of the analog trigger. True if the
        analog input is between the upper and lower limits.

        :returns: The InWindow output of the analog trigger.
        """
        if self.port is None:
            raise ValueError("operation on freed port")
        return hal.getAnalogTriggerInWindow(self.port)

    def getTriggerState(self):
        """Return the TriggerState output of the analog trigger. True if above
        upper limit. False if below lower limit. If in Hysteresis, maintain
        previous state.

        :returns: The TriggerState output of the analog trigger.
        """
        if self.port is None:
            raise ValueError("operation on freed port")
        return hal.getAnalogTriggerTriggerState(self.port)

    def createOutput(self, type):
        """Creates an :class:`.AnalogTriggerOutput` object. Gets an output object that
        can be used for routing. Caller is responsible for deleting the
        AnalogTriggerOutput object.

        :param type: An enum of the type of output object to create.
        :returns: An AnalogTriggerOutput object.
        """
        if self.port is None:
            raise ValueError("operation on freed port")
        return AnalogTriggerOutput(self, type)



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

import hal

from .counter import Counter
from .livewindow import LiveWindow

__all__ = ["GearTooth"]

class GearTooth(Counter):
    """Interface to the gear tooth sensor supplied by FIRST
    
    Currently there is no reverse sensing on the gear tooth sensor, but in
    future versions we might implement the necessary timing in the FPGA to
    sense direction.
    """

    kGearToothThreshold = 55e-6

    def enableDirectionSensing(self, directionSensitive):
        if directionSensitive:
            self.setPulseLengthMode(GearTooth.kGearToothThreshold)

    def __init__(self, channel, directionSensitive=False):
        """Construct a GearTooth sensor.

        :param channel: The DIO channel index or DigitalSource that the sensor
            is connected to.
        :type channel: int
        :param directionSensitive: True to enable the pulse length decoding in
            hardware to specify count direction.  Defaults to False.
        :type directionSensitive: bool
        """
        super().__init__(channel)
        self.enableDirectionSensing(directionSensitive)
        if hasattr(self.upSource, "getChannel"):
            if directionSensitive:
                hal.HALReport(hal.HALUsageReporting.kResourceType_GearTooth,
                              self.upSource.getChannel(), 0, "D")
            else:
                hal.HALReport(hal.HALUsageReporting.kResourceType_GearTooth,
                              self.upSource.getChannel(), 0)
        LiveWindow.addSensorChannel("GearTooth", self.upSource.getChannel(),
                                    self)

    # Live Window code, only does anything if live window is activated.

    def getSmartDashboardType(self):
        return "Gear Tooth"



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .livewindowsendable import LiveWindowSendable

__all__ = ["SensorBase"]

class SensorBase(LiveWindowSendable): # TODO: Refactor
    """Base class for all sensors
    
    Stores most recent status information as well as containing utility
    functions for checking channels and error processing.
    """

    # TODO: Move this to the HAL

    #: Ticks per microsecond
    kSystemClockTicksPerMicrosecond = 40
    
    #: Number of digital channels per roboRIO
    kDigitalChannels = 26
    
    #: Number of analog input channels
    kAnalogInputChannels = 8
    
    #: Number of analog output channels
    kAnalogOutputChannels = 2
    
    #: Number of solenoid channels per module
    kSolenoidChannels = 8
    
    #: Number of solenoid modules
    kSolenoidModules = 2
    
    #: Number of PWM channels per roboRIO
    kPwmChannels = 20
    
    #: Number of relay channels per roboRIO
    kRelayChannels = 4
    
    #: Number of power distribution channels
    kPDPChannels = 16

    #: Default solenoid module
    defaultSolenoidModule = 0

    @staticmethod
    def setDefaultSolenoidModule(moduleNumber):
        """Set the default location for the Solenoid module.

        :param moduleNumber: The number of the solenoid module to use.
        """
        SensorBase.checkSolenoidModule(moduleNumber)
        SensorBase.defaultSolenoidModule = moduleNumber

    @staticmethod
    def checkSolenoidModule(moduleNumber):
        """Verify that the solenoid module is correct.

        :param moduleNumber: The solenoid module module number to check.
        """
        pass
        #if hal.checkSolenoidModule(moduleNumber - 1) != 0:
        #    print("Solenoid module %d is not present." % moduleNumber)

    @staticmethod
    def checkDigitalChannel(channel):
        """Check that the digital channel number is valid.
        Verify that the channel number is one of the legal channel numbers.
        Channel numbers are 0-based.

        :param channel: The channel number to check.
        """
        if channel < 0 or channel >= SensorBase.kDigitalChannels:
            raise IndexError("Requested digital channel number %d is out of range." % channel)

    @staticmethod
    def checkRelayChannel(channel):
        """Check that the digital channel number is valid.
        Verify that the channel number is one of the legal channel numbers.
        Channel numbers are 0-based.

        :param channel: The channel number to check.
        """
        if channel < 0 or channel >= SensorBase.kRelayChannels:
            raise IndexError("Requested relay channel number %d is out of range." % channel)

    @staticmethod
    def checkPWMChannel(channel):
        """Check that the digital channel number is valid.
        Verify that the channel number is one of the legal channel numbers.
        Channel numbers are 0-based.

        :param channel: The channel number to check.
        """
        if channel < 0 or channel >= SensorBase.kPwmChannels:
            raise IndexError("Requested PWM channel number %d is out of range." % channel)

    @staticmethod
    def checkAnalogInputChannel(channel):
        """Check that the analog input number is value.
        Verify that the analog input number is one of the legal channel
        numbers.  Channel numbers are 0-based.

        :param channel: The channel number to check.
        """
        if channel < 0 or channel >= SensorBase.kAnalogInputChannels:
            raise IndexError("Requested analog input channel number %d is out of range." % channel)

    @staticmethod
    def checkAnalogOutputChannel(channel):
        """Check that the analog input number is value.
        Verify that the analog input number is one of the legal channel
        numbers.  Channel numbers are 0-based.

        :param channel: The channel number to check.
        """
        if channel < 0 or channel >= SensorBase.kAnalogOutputChannels:
            raise IndexError("Requested analog output channel number %d is out of range." % channel)

    @staticmethod
    def checkSolenoidChannel(channel):
        """Verify that the solenoid channel number is within limits.  Channel
        numbers are 0-based.

        :param channel: The channel number to check.
        """
        if channel < 0 or channel >= SensorBase.kSolenoidChannels:
            raise IndexError("Requested solenoid channel number %d is out of range." % channel)

    @staticmethod
    def checkPDPChannel(channel):
        """Verify that the power distribution channel number is within limits.
        Channel numbers are 0-based.

        :param channel: The channel number to check.
        """
        if channel < 0 or channel >= SensorBase.kPDPChannels:
            raise IndexError("Requested PDP channel number %d is out of range." % channel)

    @staticmethod
    def getDefaultSolenoidModule():
        """Get the number of the default solenoid module.

        :returns: The number of the default solenoid module.
        """
        return SensorBase.defaultSolenoidModule

    def free(self):
        """Free the resources used by this object"""
        # TODO: delete?
        pass



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .scheduler import Scheduler
from ..sendable import Sendable

__all__ = ["Subsystem"]

class Subsystem(Sendable):
    """This class defines a major component of the robot.

    A good example of a subsystem is the driveline, or a claw if the robot has
    one.

    All motors should be a part of a subsystem. For instance, all the wheel
    motors should be a part of some kind of "Driveline" subsystem.

    Subsystems are used within the command system as requirements for Command.
    Only one command which requires a subsystem can run at a time.  Also,
    subsystems can have default commands which are started if there is no
    command running which requires this subsystem.

    .. seealso:: :class:`.Command`
    """

    def __init__(self, name=None):
        """Creates a subsystem.

        :param name: the name of the subsystem; if None, it will be set to the
                     name to the name of the class.
        """
        # The name
        if name is None:
            self.name = self.__class__.__name__
        else:
            self.name = name
        Scheduler.getInstance().registerSubsystem(self)

        # Whether or not getDefaultCommand() was called
        self.initializedDefaultCommand = False
        # The current command
        self.currentCommand = None
        self.currentCommandChanged = True

        # The default command
        self.defaultCommand = None

    def initDefaultCommand(self):
        """Initialize the default command for a subsystem
        By default subsystems have no default command, but if they do, the
        default command is set with this method. It is called on all
        Subsystems by CommandBase in the users program after all the
        Subsystems are created.
        """
        pass

    def setDefaultCommand(self, command):
        """Sets the default command.  If this is not called or is called with
        None, then there will be no default command for the subsystem.

        :param command: the default command (or None if there should be none)
        
        .. warning:: This should NOT be called in a constructor if the subsystem
                     is a singleton.
        """
        if command is None:
            self.defaultCommand = None
        else:
            if self not in command.getRequirements():
                raise ValueError("A default command must require the subsystem")
            self.defaultCommand = command
        table = self.getTable()
        if table is not None:
            if self.defaultCommand is not None:
                table.putBoolean("hasDefault", True)
                table.putString("default", self.defaultCommand.getName())
            else:
                table.putBoolean("hasDefault", False)

    def getDefaultCommand(self):
        """Returns the default command (or None if there is none).
        
        :returns: the default command
        """
        if not self.initializedDefaultCommand:
            self.initializedDefaultCommand = True
            self.initDefaultCommand()
        return self.defaultCommand

    def setCurrentCommand(self, command):
        """Sets the current command
        
        :param command: the new current command
        """
        self.currentCommand = command
        self.currentCommandChanged = True

    def confirmCommand(self):
        """Call this to alert Subsystem that the current command is actually
        the command.  Sometimes, the Subsystem is told that it has no command
        while the Scheduler is going through the loop, only to be soon after
        given a new one.  This will avoid that situation.
        """
        if self.currentCommandChanged:
            table = self.getTable()
            if table is not None:
                if self.currentCommand is not None:
                    table.putBoolean("hasCommand", True)
                    table.putString("command", self.currentCommand.getName())
                else:
                    table.putBoolean("hasCommand", False)
            self.currentCommandChanged = False

    def getCurrentCommand(self):
        """Returns the command which currently claims this subsystem.
        
        :returns: the command which currently claims this subsystem
        """
        return self.currentCommand

    def __str__(self):
        return self.getName()

    def getName(self):
        """Returns the name of this subsystem, which is by default the class
        name.
        
        :returns: the name of this subsystem
        """
        return self.name

    def getSmartDashboardType(self):
        return "Subsystem"

    def initTable(self, table):
        super().initTable(table)
        if table is not None:
            if self.defaultCommand is not None:
                table.putBoolean("hasDefault", True)
                table.putString("default", self.defaultCommand.getName())
            else:
                table.putBoolean("hasDefault", False)

            if self.currentCommand is not None:
                table.putBoolean("hasCommand", True)
                table.putString("command", self.currentCommand.getName())
            else:
                table.putBoolean("hasCommand", False)



#----------------------------------------------------------------------------
# Copyright (c) FIRST 2008-2012. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
#----------------------------------------------------------------------------

from .pidoutput import PIDOutput

__all__ = ["SpeedController"]

class SpeedController(PIDOutput):
    """Interface for speed controlling devices."""

    def get(self):
        """Common interface for getting the current set speed of a speed
        controller.

        :returns: The current set speed.  Value is between -1.0 and 1.0.
        """
        raise NotImplementedError

    def set(self, speed, syncGroup=0):
        """Common interface for setting the speed of a speed controller.

        :param speed: The speed to set.  Value should be between -1.0 and 1.0.
        :param syncGroup: The update group to add this set() to, pending
                          updateSyncGroup().  If 0 (or unspecified), update
                          immediately.
        """
        raise NotImplementedError

    def disable(self):
        """Disable the speed controller."""
        raise NotImplementedError
